@@ -1,7 +1,6 @@
 import urllib.parse
 import requests
 
-
 class SegmentsClient:
     """SegmentsClient class.
 
@@ -19,7 +18,19 @@ class SegmentsClient:
         self.api_key = api_key
         self.api_url = api_url
 
-        r = self.get('/api_status/?lib_version=0.31')
+        # https://realpython.com/python-requests/#performance
+        # https://stackoverflow.com/questions/21371809/cleanly-setting-max-retries-on-python-requests-get-or-post-method
+        # https://stackoverflow.com/questions/23013220/max-retries-exceeded-with-url-in-requests
+        self.api_session = requests.Session()
+        adapter = requests.adapters.HTTPAdapter(max_retries=3)
+        self.api_session.mount('http://', adapter)
+        self.api_session.mount('https://', adapter)
+
+        self.s3_session = requests.Session()
+        self.s3_session.mount('http://', adapter)
+        self.s3_session.mount('https://', adapter)
+
+        r = self.get('/api_status/?lib_version=0.37')
         if r.status_code == 200:
             print('Initialized successfully.')
         elif r.status_code == 426:
@@ -65,7 +76,7 @@ class SegmentsClient:
         Args:
             name (str): The dataset name. Example: flowers.
             description (str, optional): The dataset description. Defaults to ''.
-            task_type (str, optional): The dataset's task type. Defaults to 'segmentation-bitmap'.
+            task_type (str, optional): The dataset's task type. One of 'segmentation-bitmap', 'segmentation-bitmap-highres', 'bboxes', 'keypoints'. Defaults to 'segmentation-bitmap'.
             task_attributes (dict, optional): The dataset's task attributes. Defaults to None.
             category (str, optional): The dataset category. Defaults to 'other'.
             public (bool, optional): The dataset visibility. Defaults to False.
@@ -99,6 +110,46 @@ class SegmentsClient:
         r = self.post('/user/datasets/', payload)
         return r.json()
 
+    def update_dataset(self, dataset_identifier, description=None, task_type=None, task_attributes=None, category=None, public=None, readme=None):
+        """Update a dataset.
+
+        Args:
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            description (str, optional): The dataset description.
+            task_type (str, optional): The dataset's task type. One of 'segmentation-bitmap', 'segmentation-bitmap-highres', 'bboxes', 'keypoints'.
+            task_attributes (dict, optional): The dataset's task attributes.
+            category (str, optional): The dataset category.
+            public (bool, optional): The dataset visibility.
+            readme (str, optional): The dataset readme.
+
+        Returns:
+            dict: a dictionary representing the updated dataset.
+        """
+
+        payload = {}
+
+        if description is not None:
+            payload['description'] = description
+
+        if task_type is not None:
+            payload['task_type'] = task_type
+
+        if task_attributes is not None:
+            payload['task_attributes'] = task_attributes
+
+        if category is not None:
+            payload['category'] = category
+
+        if public is not None:
+            payload['public'] = public
+
+        if readme is not None:
+            payload['readme'] = readme
+
+        r = self.patch('/datasets/{}/'.format(dataset_identifier), payload)
+        print('Updated ' + dataset_identifier)
+        return r.json()
+
     def delete_dataset(self, dataset_identifier):
         """Delete a dataset.
 
@@ -108,14 +159,33 @@ class SegmentsClient:
 
         r = self.delete('/datasets/{}/'.format(dataset_identifier))
 
+    def add_dataset_collaborator(self, dataset_identifier, username, role='labeler'):
+        """Add a collaborator to a dataset.
+
+        Args:
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            username (str): The username of the collaborator to be added.
+            role (str, optional): The role of the collaborator to be added. One of labeler, reviewer, admin. Defaults to labeler.
+
+        Returns:
+            dict: a dictionary containing the newly added collaborator with its role.
+        """
+        payload = {
+            'user': username,
+            'role': role
+        }
+        r = self.post('/datasets/{}/collaborators/'.format(dataset_identifier), payload)
+        return r.json()
+
+
     ###########
     # Samples #
     ###########
-    def get_samples(self, dataset, per_page=1000, page=1):
+    def get_samples(self, dataset_identifier, per_page=1000, page=1):
         """Get the samples in a dataset.
 
         Args:
-            dataset (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
             per_page (int, optional): Pagination parameter indicating the maximum number of samples to return. Defaults to 1000.
             page (int, optional): Pagination parameter indicating the page to return. Defaults to 1.
 
@@ -123,7 +193,7 @@ class SegmentsClient:
             list: a list of dictionaries representing the samples.
         """
 
-        r = self.get('/datasets/{}/samples/?per_page={}&page={}'.format(dataset, per_page, page))
+        r = self.get('/datasets/{}/samples/?per_page={}&page={}'.format(dataset_identifier, per_page, page))
         return r.json()
 
     def get_sample(self, uuid):
@@ -139,11 +209,11 @@ class SegmentsClient:
         r = self.get('/samples/{}/'.format(uuid))
         return r.json()
 
-    def add_sample(self, dataset, name, attributes, metadata=None, priority=None):
+    def add_sample(self, dataset_identifier, name, attributes, metadata=None, priority=None):
         """Add a sample to a dataset.
 
         Args:
-            dataset (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
             name (str): The name of the sample.
             attributes (dict): The sample attributes. Please refer to the online documentation.
             metadata (dict, optional): Any sample metadata. Example: {'weather': 'sunny', 'camera_id': 3}.
@@ -164,7 +234,7 @@ class SegmentsClient:
         if priority is not None:
             payload['priority'] = priority
 
-        r = self.post('/datasets/{}/samples/'.format(dataset), payload)
+        r = self.post('/datasets/{}/samples/'.format(dataset_identifier), payload)
         print('Added ' + name)
         return r.json()
 
@@ -234,6 +304,7 @@ class SegmentsClient:
             labelset (str): The labelset this label belongs to.
             attributes (dict): The label attributes. Please refer to the online documentation.
             label_status (str, optional): The label status. Defaults to 'PRELABELED'.
+            score (float, optional): The label score. Defaults to None.
 
         Returns:
             dict: a dictionary representing the newly created label.
@@ -258,6 +329,7 @@ class SegmentsClient:
             labelset (str): The labelset this label belongs to.
             attributes (dict): The label attributes. Please refer to the online documentation.
             label_status (str, optional): The label status. Defaults to 'PRELABELED'.
+            score (float, optional): The label score. Defaults to None.
 
         Returns:
             dict: a dictionary representing the updated label.
@@ -290,38 +362,38 @@ class SegmentsClient:
     ############
     # Releases #
     ############
-    def get_releases(self, dataset):
+    def get_releases(self, dataset_identifier):
         """Get the releases in a dataset.
 
         Args:
-            dataset (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
 
         Returns:
             list: a list of dictionaries representing the releases.
         """
 
-        r = self.get('/datasets/{}/releases/'.format(dataset))
+        r = self.get('/datasets/{}/releases/'.format(dataset_identifier))
         return r.json()
 
-    def get_release(self, dataset, release):
+    def get_release(self, dataset_identifier, name):
         """Get a release.
 
         Args:
-            dataset (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
-            release (str): The release name.
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            name (str): The name of the release.
 
         Returns:
             dict: a dictionary representing the release.
         """
 
-        r = self.get('/datasets/{}/releases/{}/'.format(dataset, release))
+        r = self.get('/datasets/{}/releases/{}/'.format(dataset_identifier, name))
         return r.json()
 
-    def add_release(self, dataset, name, description=''):
+    def add_release(self, dataset_identifier, name, description=''):
         """Add a release to a dataset.
 
         Args:
-            dataset (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
             name (str): The name of the release.
             description (str, optional): The release description.
 
@@ -333,28 +405,28 @@ class SegmentsClient:
             'name': name,
             'description': description
         }
-        r = self.post('/datasets/{}/releases/'.format(dataset), payload)
+        r = self.post('/datasets/{}/releases/'.format(dataset_identifier), payload)
         return r.json()
 
-    def delete_release(self, dataset, release):
+    def delete_release(self, dataset_identifier, name):
         """Delete a release.
 
         Args:
-            dataset (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
-            release (str): The release name.
+            dataset_identifier (str): The dataset identifier, consisting of the name of the dataset owner followed by the name of the dataset itself. Example: jane/flowers.
+            name (str): The name of the release.
         """
 
-        r = self.delete('/datasets/{}/releases/{}/'.format(dataset, release))
+        r = self.delete('/datasets/{}/releases/{}/'.format(dataset_identifier, name))
 
     ##########
     # Assets #
     ##########
-    def upload_asset(self, file, filename):
+    def upload_asset(self, file, filename='label.png'):
         """Upload an asset.
 
         Args:
             file (object): A file object.
-            filename (str): The file name.
+            filename (str, optional): The file name. Defaults to label.png.
 
         Returns:
             dict: a dictionary representing the uploaded file.
@@ -376,7 +448,7 @@ class SegmentsClient:
         headers = self._get_auth_header() if auth else None
 
         try:
-            r = requests.get(urllib.parse.urljoin(self.api_url, endpoint),
+            r = self.api_session.get(urllib.parse.urljoin(self.api_url, endpoint),
                             headers=headers)
             r.raise_for_status()
         except requests.exceptions.HTTPError as errh:
@@ -388,7 +460,7 @@ class SegmentsClient:
         headers = self._get_auth_header() if auth else None
 
         try:
-            r = requests.post(urllib.parse.urljoin(self.api_url, endpoint),
+            r = self.api_session.post(urllib.parse.urljoin(self.api_url, endpoint),
                                 json=data,  # data=data
                                 headers=headers)
             r.raise_for_status()
@@ -401,7 +473,7 @@ class SegmentsClient:
         headers = self._get_auth_header() if auth else None
 
         try:
-            r = requests.put(urllib.parse.urljoin(self.api_url, endpoint),
+            r = self.api_session.put(urllib.parse.urljoin(self.api_url, endpoint),
                                 json=data,  # data=data
                                 headers=headers)
             r.raise_for_status()
@@ -414,7 +486,7 @@ class SegmentsClient:
         headers = self._get_auth_header() if auth else None
 
         try:
-            r = requests.patch(urllib.parse.urljoin(self.api_url, endpoint),
+            r = self.api_session.patch(urllib.parse.urljoin(self.api_url, endpoint),
                                 json=data,  # data=data
                                 headers=headers)
             r.raise_for_status()
@@ -427,7 +499,7 @@ class SegmentsClient:
         headers = self._get_auth_header() if auth else None
 
         try:
-            r = requests.delete(urllib.parse.urljoin(self.api_url, endpoint),
+            r = self.api_session.delete(urllib.parse.urljoin(self.api_url, endpoint),
                                 json=data,  # data=data
                                 headers=headers)
             r.raise_for_status()
@@ -436,10 +508,9 @@ class SegmentsClient:
 
         return r
 
-    @staticmethod
-    def _upload_to_aws(file, aws_fields):
+    def _upload_to_aws(self, file, aws_fields):
         files = {'file': file}
-        r = requests.post(aws_fields['url'],
+        r = self.s3_session.post(aws_fields['url'],
                                  files=files,
                                  data=aws_fields['fields'])
         return r
\ No newline at end of file
