@@ -5,7 +5,7 @@ from urllib.parse import urlparse
 from multiprocessing.pool import ThreadPool
 from tqdm import tqdm
 
-from .utils import download_and_save_image, download_and_save_segmentation_bitmap, handle_exif_rotation
+from .utils import load_image_from_url, load_label_bitmap_from_url, handle_exif_rotation
 
 from PIL import Image
 
@@ -21,12 +21,13 @@ class SegmentsDataset():
     """
 
     # https://stackoverflow.com/questions/682504/what-is-a-clean-pythonic-way-to-have-multiple-constructors-in-python
-    def __init__(self, release_file, labelset='ground-truth', filter_by=None, segments_dir='segments'):
+    def __init__(self, release_file, labelset='ground-truth', filter_by=None, segments_dir='segments', preload=True):
         self.labelset = labelset
         self.filter_by = [filter_by] if isinstance(filter_by, str) else filter_by
         if self.filter_by is not None:
             self.filter_by = [s.lower() for s in self.filter_by]
         self.segments_dir = segments_dir
+        self.preload = preload
         
         # if urlparse(release_file).scheme in ('http', 'https'): # If it's a url
         if isinstance(release_file, str): # If it's a file path
@@ -47,14 +48,14 @@ class SegmentsDataset():
             return
 
         self.task_type = self.release['dataset']['task_type']
-        if self.task_type not in ['segmentation-bitmap', 'bboxes']:
-            print('You can only create a dataset for tasks of type "segmentation-bitmap" or "bboxes" for now.')
+        if self.task_type not in ['segmentation-bitmap', 'segmentation-bitmap-highres', 'bboxes', 'keypoints']:
+            print('You can only create a dataset for tasks of type "segmentation-bitmap", "segmentation-bitmap-highres", "bboxes", or "keypoints" for now.')
             return
         
         self.load_dataset()
 
     def load_dataset(self):
-        print('Initializing dataset. This may take a few seconds...')
+        print('Initializing dataset...')
         
         # Setup cache
         if not os.path.exists(self.image_dir):
@@ -91,9 +92,11 @@ class SegmentsDataset():
         # https://stackoverflow.com/questions/3530955/retrieve-multiple-urls-at-once-in-parallel
         # https://github.com/tqdm/tqdm/issues/484#issuecomment-461998250
         num_samples = self.__len__()
-        with ThreadPool(16) as pool:
-            # r = list(tqdm(pool.imap_unordered(self.__getitem__, range(num_samples)), total=num_samples))
-            r = list(tqdm(pool.imap_unordered(_load_image, range(num_samples)), total=num_samples))
+        if self.preload:
+            print('Preloading all samples. This may take a while...')
+            with ThreadPool(16) as pool:
+                # r = list(tqdm(pool.imap_unordered(self.__getitem__, range(num_samples)), total=num_samples))
+                r = list(tqdm(pool.imap_unordered(_load_image, range(num_samples)), total=num_samples))
 
         print('Initialized dataset with {} images.'.format(num_samples))
 
@@ -107,9 +110,10 @@ class SegmentsDataset():
         image_filename = os.path.join(self.image_dir, image_filename_rel)
 
         if not os.path.exists(image_filename):
-            download_and_save_image(image_url, image_filename)
+            image = load_image_from_url(image_url, image_filename)
+        else:
+            image = Image.open(image_filename)
 
-        image = Image.open(image_filename)
         image = handle_exif_rotation(image)
 
         return image, image_filename_rel
@@ -123,9 +127,9 @@ class SegmentsDataset():
         segmentation_bitmap_filename = os.path.join(self.image_dir, '{}_label_{}{}'.format(sample_name, labelset, url_extension))
         
         if not os.path.exists(segmentation_bitmap_filename):
-            download_and_save_segmentation_bitmap(segmentation_bitmap_url, segmentation_bitmap_filename)
-
-        segmentation_bitmap = Image.open(segmentation_bitmap_filename)
+            segmentation_bitmap = load_label_bitmap_from_url(segmentation_bitmap_url, segmentation_bitmap_filename)
+        else:
+            segmentation_bitmap = Image.open(segmentation_bitmap_filename)
 
         return segmentation_bitmap
 
@@ -158,7 +162,7 @@ class SegmentsDataset():
         }
 
         # Segmentation bitmap
-        if self.task_type == 'segmentation-bitmap':
+        if self.task_type == 'segmentation-bitmap' or self.task_type == 'segmentation-bitmap-highres':
             # Load the label
             try:
                 label = sample['labels'][self.labelset]
@@ -175,7 +179,7 @@ class SegmentsDataset():
             })
 
         # Bounding boxes
-        elif self.task_type == 'bboxes':
+        elif self.task_type == 'bboxes' or self.task_type == 'keypoints':
             try:
                 label = sample['labels'][self.labelset]
                 attributes = label['attributes']
